-- v0.2.1

local function modify_biome(xml2lua_path, biome_file_path, func)
  local file_contents = ModTextFileGetContent(biome_file_path)
  if not file_contents then
    error("File not found: " .. biome_file_path)
  end
  local xml2lua = dofile(xml2lua_path)
  local handler = xml2lua.parse(file_contents)
  local desired_changes = func(handler.root.Biome.Topology._attr)
  for k, v in pairs(desired_changes) do
    handler.root.Biome.Topology._attr[k] = v
  end
  ModTextFileSetContent(biome_file_path, xml2lua.toXml(handler.root, "Biome", 0))
end

local function find_biome_scripts(current_mod_id)
  local biome_scripts = {}
  for i, mod_id in ipairs(ModGetActiveModIDs()) do
    local init_content = ModTextFileGetContent("mods/" .. mod_id .. "/init.lua")
    if init_content then
      -- Get the magic numbers file path if present
      local magic_numbers_file_path
      local lines = init_content:gmatch("([^\r\n]+)")
      for line in lines do
        -- Make sure to only get it if it's not commented out
        local match_start, match_end, capture = line:find("ModMagicNumbersFileAdd%s*%(%s*[\"\']%s*([^%)]*)%s*[\"\']%s*%)")
        if match_start then
          if not (line:sub(1, match_start-1):find("%-%-")) then
            magic_numbers_file_path = capture
          end
        end
      end
      -- Try to get the BIOME_MAP path
      local biome_map_script_path
      if magic_numbers_file_path then
        local magic_numbers_content = ModTextFileGetContent(magic_numbers_file_path)
        if magic_numbers_content then
          local xml2lua = dofile("mods/" .. current_mod_id .. "/lib/xml2lua.lua")
          local handler = xml2lua.parse(magic_numbers_content)
          biome_map_script_path = handler.root.MagicNumbers._attr.BIOME_MAP
        end
      end
  
      if biome_map_script_path then
        table.insert(biome_scripts, { path = biome_map_script_path, mod_id = mod_id })
      end
    end
  end
  return biome_scripts
end

return function(mod_id)
  -- Only run this once globally
  if not ModTextFileGetContent("data/_biome_utils_generated_magic_numbers.xml") then
    local biome_scripts = find_biome_scripts(mod_id)
    for i, script in ipairs(biome_scripts) do
      print("BIOME SCRIPT FOUND: " .. tostring(script.mod_id) .. tostring(script.path))
      -- So this script doesn't get called twice, because we also append the assembler to the nightmare script itself
      if script.mod_id ~= "nightmare" then
        ModLuaFileAppend(script.path, string.format("mods/%s/lib/biome_utils_map_assembler.lua", mod_id))
      end
    end
    ModTextFileSetContent("data/biome_utils_map_assembler_nightmare.lua", string.format([[
      BIOME_UTILS_MAP_TYPE = "NIGHTMARE"
      loadfile("mods/%s/lib/biome_utils_map_assembler.lua")()
    ]], mod_id))
    ModLuaFileAppend("mods/nightmare/files/biome_map.lua", "data/biome_utils_map_assembler_nightmare.lua")

    ModTextFileSetContent("data/biome_utils_map_assembler_ngplus_map_script.lua", string.format([[
      BIOME_UTILS_MAP_TYPE = "NGPLUS"
      loadfile("mods/%s/lib/biome_utils_map_assembler.lua")()
    ]], mod_id))
    ModLuaFileAppend("data/biome_impl/biome_map_newgame_plus.lua", "data/biome_utils_map_assembler_ngplus_map_script.lua")
  end

  local temp_magic_numbers_filepath = "data/_biome_utils_generated_magic_numbers.xml"
  if not ModTextFileGetContent(temp_magic_numbers_filepath) then
    ModTextFileSetContent(temp_magic_numbers_filepath, string.format([[<MagicNumbers BIOME_MAP="mods/%s/lib/biome_utils_map_assembler.lua" /> ]], mod_id))
    ModMagicNumbersFileAdd(temp_magic_numbers_filepath)
  end

  local function set_map_script(map_script_path, map_type, priority)
    priority = priority or 0
    map_type = map_type and ("_" .. map_type) or ""
    local map_script_storage_path = "data/biome_compat_registered_map_scripts" .. map_type .. ".txt"
    local content = ModTextFileGetContent(map_script_storage_path) or ""
    assert(type(priority) == "number", "priority must be a number")
    content = content .. mod_id .. ":" .. map_script_path .. ":" .. priority .. "\n"
    ModTextFileSetContent(map_script_storage_path, content)
  end

  return setmetatable({
    register_map_script = function(map_script_path, priority) set_map_script(map_script_path) end,
    register_map_script_nightmare = function(map_script_path, priority) set_map_script(map_script_path, "nightmare") end,
    register_map_script_ngplus = function(map_script_path, priority) set_map_script(map_script_path, "ngplus") end,
    add_biomes = function(...)
      -- Make sure the function can be used with varargs or a table
      local varargs = ...
      assert(type(varargs) == "table", "Incorrect argument format")
      if #varargs[1] == 3 and type(varargs[1][1]) == "string" and type(varargs[1][2]) == "number" and type(varargs[1][3]) == "string" then
      else
        varargs = {...}
      end
      local biomes_to_add = varargs
      
      -- biome_filename, height_index, color
      local biomes_all_xml = ModTextFileGetContent("data/biome/_biomes_all.xml")
      local xml2lua = dofile("mods/" .. mod_id .. "/lib/xml2lua.lua")
      local handler = xml2lua.parse(biomes_all_xml)
    
      for i, biome_to_add in ipairs(biomes_to_add) do
        table.insert(handler.root.BiomesToLoad.Biome, { _attr = {
          biome_filename=biome_to_add[1],
          height_index=biome_to_add[2],
          color=biome_to_add[3],
        }})
        print("Inserting biome: " .. biome_to_add[1] .. ", " ..biome_to_add[3])
      end
    
      local output_xml_content = xml2lua.toXml(handler.root, "BiomesToLoad", 0)
      ModTextFileSetContent("data/biome/_biomes_all.xml", output_xml_content)
      return output_xml_content
    end,
    modify_biome = function(biome_file_path, func)
      modify_biome("mods/" .. mod_id .. "/lib/xml2lua.lua", biome_file_path, func)
    end,
  }, {
    __newindex = function(t, prop, value)
      if prop == "biome_offset_y" then
        local exists = ModTextFileGetContent("data/biome_utils_biome_offset_y.lua")
        if exists then
          local func, err = loadfile("data/biome_utils_biome_offset_y.lua")
          local mod_id, offset_y_set_to = unpack(func())
          if tonumber(offset_y_set_to) ~= tonumber(value) then
            error("biome_offset_y has already been set to " .. offset_y_set_to .. " by mod with id \"" .. mod_id .. "\"!")
          end
        else
          local biomes_all_xml = ModTextFileGetContent("data/biome/_biomes_all.xml")
          local xml2lua = dofile("mods/" .. mod_id .. "/lib/xml2lua.lua")
          local handler = xml2lua.parse(biomes_all_xml)
          handler.root.BiomesToLoad._attr.biome_offset_y = value
          local output_xml_content = xml2lua.toXml(handler.root, "BiomesToLoad", 0)
          ModTextFileSetContent("data/biome/_biomes_all.xml", output_xml_content)
          ModTextFileSetContent("data/biome_utils_biome_offset_y.lua", string.format("return { \"%s\", %s }", mod_id, value))
        end
      end
    end
  })
end
