-- v0.2.0

local function extract_argb(argb)
  local a = bit.rshift(bit.band(argb, 0xff000000), 3 * 8)
  local r = bit.rshift(bit.band(argb, 0x00ff0000), 2 * 8)
  local g = bit.rshift(bit.band(argb, 0x0000ff00), 1 * 8)
  local b = bit.band(argb, 0x000000ff)
  return a, r, g, b
end

local function split_string(inputstr, sep)
  sep = sep or "%s"
  local t= {}
  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
    table.insert(t, str)
  end
  return t
end

local function table_get_keys(t)
  local keys = {}
  for k, v in pairs(t) do table.insert(keys, k) end
  return keys
end
-- Store x, y as one number by putting x in the lower 2 bytes and y in the upper bytes
local function encode_coords(x, y)
  return x + y * 0x10000
end

local function decode_coords(coords)
  return bit.band(coords, 0xffff), bit.rshift(coords, 16)
end

local incompat_found_message = "------------------ !! BIOME MOD INCOMPATIBILITY DETECTED !! ------------------"
local base_map_path = nil
local base_map_size_x = nil
local base_map_size_y = nil
local aggregated_pixeldata = {}

local _BiomeMapSetSize = BiomeMapSetSize
local _BiomeMapGetPixel = BiomeMapGetPixel
local _BiomeMapSetPixel = BiomeMapSetPixel
local _BiomeMapLoadImage = BiomeMapLoadImage

-- Sets overlapping mismatching pixels to black and prints warnings to the console (priority is unused atm)
function BiomeMapCheckForIncompatibilities(orig_priority)
  local mismatches = {}
  local mismatch_found = false
  for coords, data in pairs(aggregated_pixeldata) do
    local x, y = decode_coords(coords)
    local temp_color = 0
    for i, v in ipairs(data) do
      if temp_color ~= 0 and v.color ~= temp_color then
        mismatches[coords] = data
        mismatch_found = true
      end
      temp_color = v.color
    end
  end
  if mismatch_found then
    print(incompat_found_message)
  end
  for coords, data in pairs(mismatches) do
    local x, y = decode_coords(coords)
    print("Biome map script priority: " .. orig_priority)
    print("Pixel at " .. x .. ", " .. y .. " is being set to different colors by multiple mods:")
    for i, v in ipairs(data) do
      print(v.mod_id .. " sets it to: " .. string.format("%08X", v.color))
    end
    -- Set the mismatching pixels to black so mod developers can better see where it is in Noita/temptemp/biome_map_lua.png
    data[1].color = 0xFF000000
  end
end

function BiomeMapAssemble()
  if not base_map_path then error("No biome map loaded! BiomeMapLoadImage was never called.") end
  print("Assembling biome map")
  print("Loading biome image: " .. tostring(base_map_path))
  _BiomeMapLoadImage(x, y, base_map_path)
  for coords, data in pairs(aggregated_pixeldata) do
    local x, y = decode_coords(coords)
    _BiomeMapSetPixel(x, y, data[1].color)
  end
end

function run_map_script_in_environment(path, mod_id, priority)
  function BiomeMapSetSize(x, y)
    if (base_map_size_x and base_map_size_x ~= x) or (base_map_size_y and base_map_size_y ~= y) then
      error("Different map sizes defined.", 2)
    end
    base_map_size_x = x
    base_map_size_y = y
    _BiomeMapSetSize(x, y)
  end
  -- Returns the pixel of the loaded biome map at x, y or the overlaid pixel if any have been set
  -- the output color in the same format that SetPixel uses, no more conversion necessary
  function BiomeMapGetPixel(x, y)
    local coords = encode_coords(x, y)
    if aggregated_pixeldata[coords] then
      -- TODO: Check if there are more than 1 (priority)?
      return aggregated_pixeldata[coords][1].color
    else
      return BiomeMapConvertPixelFromUintToInt(_BiomeMapGetPixel(x, y))
    end
  end
  
  function BiomeMapSetPixel(x, y, color, path)
    local coords = encode_coords(x, y)
    local a, r, g, b = extract_argb(color)
    aggregated_pixeldata[coords] = aggregated_pixeldata[coords] or {}
    table.insert(aggregated_pixeldata[coords], { color = color, mod_id = mod_id, image_path = path })
  end

  -- Doesn't actualy load an image yet, just stores the desired path for later assembly
  function BiomeMapLoadImage(x, y, path)
    if not base_map_path then
      base_map_path = path
    elseif base_map_path ~= path then
      print(incompat_found_message)
      print(string.format("BiomeMapLoadImage called with different paths: %s, %s", base_map_path, path))
      -- TODO: Check if they are the same image then allow them anyways?
    end
  end
  
  -- Reads out the pixels of an image and stores each non black pixel in a table for assembling later
  function BiomeMapOverlayImage(path, dest_x, dest_y, src_x1, src_y1, src_x2, src_y2)
    -- Read out all pixels and store them for lookup
    -- For that we need to load the map first though
    _BiomeMapLoadImage(0, 0, path)
    for y=src_y1, src_y2-1 do
      for x=src_x1, src_x2-1 do
        local pixel_color = BiomeMapConvertPixelFromUintToInt(_BiomeMapGetPixel(x, y))
        -- Ignore transparent pixels
        local a, r, g, b = extract_argb(pixel_color)
        if a > 0 then
          BiomeMapSetPixel(dest_x + x, dest_y + y, pixel_color, path)
        end
      end
    end
  end
  dofile(path)
end

local map_scripts = {}
local content

if not BIOME_UTILS_MAP_TYPE then
  content = ModTextFileGetContent("data/biome_compat_registered_map_scripts.txt")
elseif BIOME_UTILS_MAP_TYPE == "NGPLUS" then
  content = ModTextFileGetContent("data/biome_compat_registered_map_scripts_ngplus.txt")
elseif  BIOME_UTILS_MAP_TYPE == "NIGHTMARE" then
  content = ModTextFileGetContent("data/biome_compat_registered_map_scripts_nightmare.txt")
end
-- Check if file exists
if content then
  -- Each script is placed on a new line with the format:
  -- mod_id:script/path.lua:5
  -- where 5 would be the priority
  local lines = split_string(content, "\n")
  for i, v in ipairs(lines) do
    local mod_id, script_path, priority = unpack(split_string(v, ":"))
    map_scripts[priority] = map_scripts[priority] or {}
    table.insert(map_scripts[priority], { mod_id = mod_id, script_path = script_path })
  end
end
-- Sort all map scripts into different arrays depending on priority and run their scripts in a fake environment which collects all calls
local keys = table_get_keys(map_scripts)
table.sort(keys)
for _, orig_priority in ipairs(keys) do
  for priority, map_script in ipairs(map_scripts[orig_priority]) do
    print("Running map script: " .. map_script.script_path .. " (" .. map_script.mod_id .. ")")
    run_map_script_in_environment(map_script.script_path, map_script.mod_id, priority)
    BiomeMapCheckForIncompatibilities(orig_priority)
  end
end

BiomeMapAssemble()

-- Store the data in a file so other mods can read the final map data out if they want
local content = "return {\n"
content = content .. "  width = " .. base_map_size_x .. ",\n"
content = content .. "  height = " .. base_map_size_y .. ",\n"
content = content .. "  map = {\n"
for y=0, base_map_size_y-1 do
  for x=0, base_map_size_x-1 do
    local color = BiomeMapConvertPixelFromUintToInt(_BiomeMapGetPixel(x, y))
    content = content .. string.format([[    ["%s_%s"] = { color_hex = "%s", color_int = %s },]] .. "\n", x, y, string.format("%08X", color), color)
  end
end
content = content .. "} }\n"
ModTextFileSetContent("data/biome_utils_map_data.lua", content)

function encode_coords2(x, y)
  return (x + 0x7FFF) + (y + 0x7FFF) * 0x10000
end

local function _BiomeMapGetPixel(x, y)
  local abgr = BiomeMapGetPixel(x, y)
  local b = bit.rshift(bit.band(abgr, 0xff0000), 2 * 8)
  local g = bit.rshift(bit.band(abgr, 0x00ff00), 1 * 8)
  local r = bit.band(abgr, 0x0000ff)
  return b, g, r
end

local content = "return {"
content = content .. "width = " .. base_map_size_x .. ","
content = content .. "height = " .. base_map_size_y .. ","
content = content .. "map = {"
for y=0,base_map_size_y-1 do
  for x=0,base_map_size_x-1 do
    local r, g, b = _BiomeMapGetPixel(x, y)
    content = content .. "["..encode_coords2(x, y).."] = " .. string.format("{ r = %s, g = %s, b = %s },\n", r, g, b)
  end
end

content = content .. "}}\n"
ModTextFileSetContent("mods/LocationTracker/_virtual/map.lua", content)
GameAddFlagRun("locationtracker_reload_map")
